/*********************************************************************

SIMWORLD                                                
Version 4.0 Beta

Basic Agents

Copyright (C) 2002 Matthias Scheutz (mscheutz@nd.edu)
Artificial Intelligence and Robotics Laboratory
Department of Computer Science and Engineering
University of Notre Dame, USA
http://www.nd.edu/~mscheutz/
http://www.nd.edu/~airolab/
All rights reserved.

This program is licensed and distributed under the terms stated
in the accompanying COPYRIGHT file.

Last modified: 11-27-02

*********************************************************************/
compile_mode :pop11 +strict;

lconstant verbose = false;

;;; rulesystems for the basic agent, will be defined later
define :rulesystem basic_agent_rulesystem; enddefine;


;;;-------------------- basic (inanimate) things -----------------------
;;; defines the base class for all animate and inanimate things
;;; at a location within the simulation, with a mass and a shape
define :class thing; is sim_object;
    ;;; write over the sim_agent slot which generates all kinds of symbols
    ;;; unnecessarily ...
    slot sim_name == false;
    slot sim_name_base == false;
    ;;; geometry description of object, for now circle == false otherwise
    ;;; convex polygon
    slot shape == false;
    ;;; distance from center to boundary, radius for circles
    ;;; a vector of boundary points for a convex polygon
    slot min_dist == 5;
    slot sim_x = false;
    slot sim_y = false;
;;; slot sim_z;
;;;    slot mass = undef;
    slot bodycolor == 'black';
    slot updated == false;
    ;;; things have no gender
    slot gender == false;
enddefine;

;;; obstacles are just red inanimate things
define :class obstacle; is thing;
    ;;; things have no inner life, no internal processing mechanisms
    ;;; and produce no sounds
    slot sounditems == [];
    slot sim_rulesystem == [];
    slot sim_sensors == [];
    slot bodycolor == 'red';
    slot shape = {{-9 -9} {-9 9} {9 9} {9 -9}};
enddefine;

;;; moving obstacles are (static) obstacles that move
define :class moving_obstacle; is obstacle;
    ;;; moves every so many cycles by speed in the given heading
    slot heading = random(360);
    slot sim_interval = 2 + random(2);
    slot speed = 2 + random(2);
enddefine;

;;; define a generic parent for water and plants
define :mixin food; is thing;
    slot sounditems == [];
    slot sim_rulesystem == [];
    slot sim_sensors == [];
enddefine;

;;; define a class for plants, which eventually can grow and change their
;;; energy over time
define :class plant; is food;
    slot energy == 800;          ;;; default energy
    slot remaining_time == 200;  ;;; before the plant disappears
    slot bodycolor == 'green';
enddefine;

;;; water
define :class water; is food;
    slot energy == 650;          ;;; default energy
    slot remaining_time == 1000; ;;; before water disappears
    slot bodycolor == 'lightblue';
enddefine;


;;; ----------------- the basic agents -------------------------------

;;; defined for sonar, smell, touch, sound, vision, taste
define :class external_agent_sensors;
    ;;; sensor limits for external sensor
    ;;; low level sensory information
    slot sonar_range == 100;     ;;; includes enemies and obstacles
    slot smell_range == 300;     ;;; includes water, food, enemies and mates
    slot vision_range == 200;    ;;; for things in front of the agent
    slot vision_degrees == 60;   ;;; +- degrees for range
    slot sound_range == 100;     ;;; for sound around the agent
    slot touch_range == 35;       ;;; for things in front of the agent
    slot touch_degrees == 60;    ;;; +- degrees for touch
    slot taste_range == 4;       ;;; has to be in the mouth

    ;;; FOR BOOK KEEPING, PUT MAX RANGE IN HERE...
    slot max_range == 300;

    ;;; higher level sensory information
    ;;; different smells are separated out and associated with values

    ;;; the respective vector fields
    slot foodvector = {%0,0%};    ;;; current food vector   
    slot watervector = {%0,0%};   ;;; current water vector  
    slot agentvector = {%0,0%};   ;;; current agent vector  
    slot samevector = {%0,0%};    ;;; current same vector   
    slot othervector = {%0,0%};   ;;; current agent vector  
    slot obstvector = {%0,0%};    ;;; current obstacle vector

    slot movevector = {%0,0%};    ;;; this holds the combined vector

    ;;; higher perceptual mechanisms, object perception
    slot visionitems == [];      ;;; report individual objects
    slot sounditems == [];       ;;; report individual words
    slot touchitems == [];       ;;; report whether obstacles, agents or food/water
enddefine;

;;; defined for hands, feet, head, voice, sexual organ
define :class external_agent_effectors;
    ;;; effector range
    slot ingesting_range == 5;   ;;; for ingesting things and inflicting pain on others
enddefine;

;;; define the complete body of the agent including how it is visualized
define :class agent_body;
    slot min_dist == 9;  		 ;;; default body size
    slot bodycolor == 'brown';   ;;; default is brown
    slot gender == false;        ;;; default is no gender, otherwise it could be "male" or "female"
    slot age == 0;               ;;; in cycles
    slot createdat == 0;         ;;; when the agent was born
    slot parentID == false;      ;;; name of the parent
    slot numoffspring == 0;      ;;; the number of offspring the agent produced
    slot procreation_age = OMEGA;  ;;; in cycles
    slot procreation_latency == 0; ;;; procreation interval
    slot fertility_period = FERT;  ;;; the period during which the agent can have offspring
    slot procreation_energy == 2200;  ;;; min enery needed for procreation
    slot procreation_coolant == 1600;  ;;; coolant needed for one offspring
    slot heading = random(360);   ;;; random initial heading
    slot speed == 3;             ;;; also used as energy cost for now...
    slot maxspeed == 10;         ;;; might depend on health, etc.
    slot startupenergy == 2000;  ;;; the energy the agent starts out with
    slot startupwater == 1500;   ;;; the water the agents starts out with
    slot energy == 2000;         ;;; needs to be the same as the startup value
    slot coolant == 1500;        ;;; needs to be the same as the startup value
    slot maxenergy == 8000;
    slot maxcoolant == 6000;
    slot critcoolant == 300;     ;;; these are values for the food and water
    slot critenergy == 400;      ;;; alarms to ignore anything but food and water

    ;;; the basic alarm mechanisms (i.e., reflex for obstacles and agents)
    slot coolant_alarm == false;
    slot energy_alarm == false;
    slot agent_alarm == false;
    slot same_alarm == false;
    slot other_alarm == false;
    slot obst_alarm == false;        ;;; stores the activation of obstacle alaram
    slot reflexspeed = 7;       ;;; the speed at which the obstacle reflex is performed
    slot immanent_collision_range == 35;

    slot maxfoodintake == 50;    ;;; maximum amount of food that can be consumed in one cycle
    slot maxwaterintake == 50;   ;;; maximum amount of water that can be consumed in one cycle

    ;;; for book keeping purposes...
    slot fooditem == false;
    slot wateritem == false;
enddefine;

;;; variables of the basic schema-based agent controller
define :class gain_values;
    ;;; to keep track of the original values
    slot base_obst_factor == -50;     ;;; usually negative
    slot base_food_factor == 20;     ;;; usually positive
    slot base_water_factor == 10;     ;;; usually positive
    slot base_agent_factor == -20;
    slot base_same_factor == -20;
    slot base_other_factor == -5;

    slot food_factor == 20;      ;;; usually positive
    slot water_factor == 10;     ;;; usually positive
    slot obst_factor == -20;     ;;; usually negative
    slot agent_factor == -20;
    slot same_factor == -20;
    slot other_factor == -5;
enddefine;

;;; the basic agent
define :class basic_agent; is agent_body,
                              gain_values,
                              external_agent_sensors,
                              external_agent_effectors,
			      thing
                           ;

    ;;; this will determine, what the agent feeds on
    ;;; right now they feed on water and plants (but are food!)
    slot feedson = [%isfood%];       ;;;

    ;;; Default set of rulesets to be obeyed. Rules defined below.
    slot sim_rulesystem = basic_agent_rulesystem;
enddefine;

;;;---------------------------------------------------------------------
;;; Methods for the thing classes and agents
;;;---------------------------------------------------------------------


;;;---------------  redefined functions for basic sim objects ----------

;;; computes and returns distance between two objects
define :method sim_distance(o1:sim_object,o2:sim_object);
    sqrt(((o1.sim_x - o2.sim_x).dup *) + ((o1.sim_y - o2.sim_y).dup *))
enddefine;

define sim_distance_from(x1, y1, x2, y2) /* -> dist */;
        sqrt(((x2 - x1).dup *) + ((y2 - y1).dup *)) /* -> dist */
enddefine;

;;; MS: just so we won't get an unnecessary declaration, will be defined
;;; after new object
define :method initialize(obj:thing,parents);
enddefine;

;;; ------------------------ creation methods --------------------------
;;; this is the new generic object creation method, which will also set
;;; instance slots for the newly created object
;;; it requires a global property "constructor", where the constructor for the
;;; class named by "kind" is stored;  this can either take only a word, then
;;; the respective constructor is lookup and applied if it exists, otherwise
;;; false is returned if it does not exist;
define newobject(constructor,parents,kind /*, initials*/ ) -> object;
    lvars initials = [];
    ;;; check if initials were passed too and extract them
    if islist(kind) then
	(constructor, parents, kind) -> (constructor, parents, kind, initials)
    endif;

    lvars object = if isproperty(constructor) then constructor(kind)()
 	           elseif isprocedure(constructor) then constructor(kind)
                   else false endif;

    ;;; Initializing slots
    unless null(initials) then
	lvars i;
	for i in initials do
	    lvars (field,value) = explode(i);
	    if field == "sim_name" then
		value -> sim_name_base(object);
	    else
		value -> valof(field)(object);
	    endif;
	endfor;
    endunless;
    if object.sim_name_base == false then
	kind -> object.sim_name_base;
    endif;
    if object then
	initialize(object,parents);
    endif;
enddefine;

;;; these are prototypes that will be redefined during simulation startup with
;;; user-specified functions

;;; parents can be a single parent or a list of parents
;;; this needs to take into account that parents could be of a different kind if
;;; transformation is used in offspring...
;;; standardly: create random x, y location within world and new name
define :method initialize(obj:thing,parents);
    ;;; first create a name, but don't make it refer to the object
    if null(parents) then
    ;;;if parents = [] then
	unless isnumber(obj.sim_x) and isnumber(obj.sim_y) then
	    random_xy()->(obj.sim_x,obj.sim_y);
	endunless;
    else
	parents.sim_name_base -> obj.sim_name_base;
    endif;
    gensym(sim_name_base(obj))->obj.sim_name;
enddefine;

;;; defines whether an agent can procreate, usually things don't procreate
define :method readytoprocreate(obj:thing) -> ready;
    false -> ready;
enddefine;

;;; defines whether an agent can procreate
;;; returns true or false
define :method readytoprocreate(agent:basic_agent) -> ready;
    lvars pro = procreation_age(agent);
    ;;; if procreation_age is <false>, then the agent can never procreate
    returnunless(pro)(false);

    lvars ag = age(agent), fer = fertility_period(agent), pro = procreation_age(agent);
    ;;; if fertility_period is <false>, then there is no such period
    age(agent) > pro and
    procreation_latency(agent) = 0 and   ;;; make sure the agent is ready again...
    energy(agent) > procreation_energy(agent) and
    coolant(agent) > procreation_coolant(agent) and
    (if fer then if ag < pro+fer then true else false->>procreation_age(agent) endif else true endif)
      -> ready;
enddefine;

;;; this defines what happens when the agent has an offspring
define :method offspring(obj:thing) -> newthing;
enddefine;

;;; this defines what happens when the agent has an offspring
;;; if transformation is used, it should go in here and override this method
define :method offspring(entity:basic_agent) -> newthing;
    ;;; subtract the cost for giving birth...
    energy(entity) - procreation_energy(entity) -> energy(entity);
    coolant(entity) - procreation_coolant(entity) -> coolant(entity);

    ;;; increase offspring count in partent
    numoffspring(entity) + 1 -> numoffspring(entity);

    ;;; create an offspring of the kind of "obj"
    ;;; SHOULD THERE BE MORE INITIALIZATION???
;;;npr('Giving birth :' >< isbasic_agentGUI(entity) >< ' kind yields: ' >< kind(entity));

    newobject(constructors,entity,kind(entity)) -> newthing;
enddefine;

;;; initialization for the basic agent with one parent
define :method initialize(agent:basic_agent,parents);
    ;;; check if single parent
    unless islist(parents) then
	parents.sim_name_base -> agent.sim_name_base;
	gensym(sim_name_base(agent))->agent.sim_name;
	random_close_to(sim_x(parents),sim_y(parents),4*min_dist(parents))
	    -> (agent.sim_x,agent.sim_y);

	;;; inherit the rulesystem from the parent
    	sim_original_rulesystem(parents) -> sim_rulesystem(agent);

    	;;; set the parent ID for the new agent
    	sim_name(parents) -> parentID(agent);
    ;;; check if no parent
    elseif null(parents) then
	;;; then call the general initialization for agents
	call_next_method(agent,parents);
	valof(consword(word_string(sim_name_base(agent)) >< '_rulesystem'))
	    -> sim_rulesystem(agent);
	;;;basic_agent_rulesystem -> sim_rulesystem(agent);
    ;;; crossover here...
    else
    endunless;
enddefine;

;;; for mutation, usually things don't mutate
define :method mutate(obj:thing);
enddefine;

;;; ------------------- MUTATION -------------------------------
;;; mutates a vector or a two dimensional array "mut" according to a masking
;;; vector or array "ok", which contains "true" for the components that should
;;; be mutated at a rate of "rate" using +/- of the amount of "strength"
define mutatestruc(mut,ok,rate,strength);
    if isvector(mut) then
        lvars i, l = length(mut);
        for i from 1 to l do
            if ok(i) then
                if (random(rate)>rate-1) then
                    mut(i) - 2*strength + random(3)*strength -> mut(i);
                endif;
            endif;
        endfor;
    elseif isarray(mut) then
        lvars i,j, (n, l, m, k) = explode(boundslist(mut));
        for i from n to l do
            for j from m to k do
                if ok(i,j) then
                    if (random(rate)>rate-1) then
                        mut(i,j) - 2*strength + random(3)*strength  -> mut(i,j);
                    endif;
                endif;
            endfor;
         endfor;
    endif;
enddefine;

;;; mutates a vector or a two dimensional array "mut" by applying mproc to it
;;; the procedure must take one argument and return the mutated value
define mutatewithproc(mut,mproc);
    lvars procedure mproc;
    if isvector(mut) then
        lvars i, l = length(mut);
        for i from 1 to l do
			mproc(mut(i)) -> mut(i);
        endfor;
    elseif isarray(mut) then
        lvars i,j, (n, l, m, k) = explode(boundslist(mut));
        for i from n to l do
            for j from m to k do
                mproc(mut(i,j)) -> mut(i,j);
            endfor;
         endfor;
    endif;
enddefine;


;;; for statistics, use cycle to determine when to record
define :method record(obj:thing,cycle);
enddefine;

;;; ------------------------ other methods ---------------------------

;;; The run sensor and agent methods...
define :method sim_run_sensors(obj:thing, agents) -> sensor_data;
    ;;; things don't sense anything
    [] -> sensor_data;
enddefine;

;;; for moving obstacles, override sim_run_agent with simpler version
define :method sim_run_agent(obst:moving_obstacle, objects);
    returnunless((sim_cycle_number mod sim_interval(obst)) == 0);

    lvars x = sim_x(obst), y = sim_y(obst),
         h = heading(obst);

    ;;; TODO: MAKE THIS CONSISTENT WITH SIMULATOR.P and THE GUI VERSION
    if x > world_lim_x or x < -world_lim_x then
        if h < 180 then
            180 - h -> h
        else
            540 - h -> h
        endif
    elseif y > world_lim_y or y < -world_lim_y then
        360 - h -> h
    endif;
    h -> heading(obst);
    x + speed(obst) * cos(h) -> x;
    y + speed(obst) * sin(h) -> y;
    x -> sim_x(obst);
    y -> sim_y(obst);
enddefine;

;;; this removes an arbitrary entity from the simulation
define :method dies_entity(entity:thing);
    ;;; mark food item as dead and undraw
    "dead" -> sim_status(entity);
    conspair(entity,sim_object_delete_list) -> sim_object_delete_list;
enddefine;

;;; this is for regular updates of an entity
define :method update_body(entity:thing);
enddefine;

;;; food (i.e., plants and water) standardly ages
define :method update_body(entity:food);
    if (remaining_time(entity) < 1) then
        dies_entity(entity);
    else
        remaining_time(entity) - 1 -> remaining_time(entity);
    endif;
    true -> updated(entity);
enddefine;

;;;---------------------------------------------------------------------
;;; Methods for the agent classes
;;;---------------------------------------------------------------------

;;; ------------------------ sensor methods -----------------------------

;;; from sim_geom.p
define sim_angle_xy(x, y) -> degrees  ;
        ;;; given coordinates of point work out angle, between 0 and 360
        dlocal popradians = false;
        arctan2(x,y) -> degrees;
        if degrees < 0 then
                degrees + 360 -> degrees
        endif
enddefine;

define sim_heading_from(x1, y1, x2, y2) /* -> heading */;
        sim_angle_xy(x2 - x1, y2 - y1) /* -> heading */
enddefine;

;;; helper functions for constructing the sensor vector field
;;; gets the adjusted force vector...
define get_force(from_x, from_y, to_x, to_y, factor) -> (x, y);
    lvars dx = to_x - from_x,
          dy = to_y - from_y,
          distance = 0.0001 + sqrt(dx * dx + dy * dy),    ;;; to make it != 0
          force = factor / (distance * distance);

    dx * force -> x;
    dy * force -> y
enddefine;

define get_force3(from_x, from_y, to_x, to_y, factor) -> (x, y);
    lvars dx = to_x - from_x,
          dy = to_y - from_y,
          distance = 0.0001 + sqrt(dx * dx + dy * dy),    ;;; to make it != 0
          force = factor / (distance * distance * distance);

    dx * force -> x;
    dy * force -> y
enddefine;


;;; used for checking if something is in visual range
define in_range(dist,range,dir1,dir2,dev);
    if dir1 < dev and dist <= range then
        dir2 <= dir1 + dev or 360 - (dev - dir1) <= dir2
    elseif dir1 > 360 - dev and dist <= range then
        dir1 - dev <= dir2 or dir2 <= dir1 + dev - 360
    else
        dist <= range and dir1 - dev <= dir2 and dir2 <= dir1 + dev
    endif;
enddefine;

;;; determines what the agent feeds on, returns true or false
define :method isfoodforagent(agent:basic_agent,entity)->res;
    lvars item, res = true;
    for item in feedson(agent) do
        returnif(item(entity));
    endfor;
    false -> res;
enddefine;

;;; the agent dies, remove it from the simulation
define :method dies_basic_agent(agent:basic_agent,cause);
    ;;; make sure the agent won't die twice
    unless (sim_status(agent) == "dead") then
        "dead" -> sim_status(agent);
        conspair(agent,sim_object_delete_list) -> sim_object_delete_list;

	if KEEPAGENTS then
	    conspair(agent,dead_agents) -> dead_agents;
    	endif;

        if verbose then
            npr(agent >< ' dies at age ' >< age(agent) >< ', cause: ' >< cause);
        endif;

        ;;; gather various statistics
        lvars k = class_name(datakey(agent));
        agent_age(k) + age(agent) -> agent_age(k);
        if cause == "collision" then
	    agent_crash(k) + 1 -> agent_crash(k);
        elseif cause == "hunger" then
            agent_hunger(k) + 1 -> agent_hunger(k);
	elseif cause == "thirst" then
	    agent_thirst(k) + 1 -> agent_thirst(k);
 	elseif cause == "eaten" then
	    agent_eaten(k) + 1 -> agent_eaten(k);
    	endif;
    endunless;
enddefine;

;;; These determine whether affective agents can sense the affective state
;;; of other affective agents and react accordingly
vars affectsensethresh = 2.0;

;;; redefined sensor function for basic agent
define :method sim_run_sensors(agent:basic_agent, entities) -> sensor_data;
    lvars sensor_data = [];
    ;;; return if the agent is already marked dead...
    returnif(sim_status(agent) == "dead");

    ;;; assume there is no item in range
    false -> fooditem(agent);
    false -> wateritem(agent);
    false -> updated(agent);

    ;;; local variables to keep track of vectors, etc.
    lvars entity, sonar_rng = sonar_range(agent),
                  smell_rng = smell_range(agent),
                  vision_rng = vision_range(agent),
                  sound_rng = sound_range(agent),
                  touch_rng = touch_range(agent),
                  ingesting_rng = ingesting_range(agent),
                  obstx = 0, obsty = 0,
                  foodx = 0, foody = 0,
                  waterx = 0, watery = 0,
                  agentx = 0, agenty = 0,
                  otherx = 0, othery = 0,
                  samex = 0, samey = 0,
                  snditems = [],
                  visionitems = [],
                  touchitems = [],
                  fooditems = [],
                  obstcolliding = false,
                  agentcolliding = false,
                  sameagentcolliding = false,
                  othercolliding = false,
                  foodfactor = food_factor(agent),
                  waterfactor = water_factor(agent),
                  obstfactor = obst_factor(agent),
  	              agentfactor = agent_factor(agent),
				  otherfactor = other_factor(agent),
				  samefactor = same_factor(agent),
                  ax = sim_x(agent),
                  ay = sim_y(agent),
                  ah = heading(agent),
                  maxrange = max_range(agent);

    ;;; compute inputs from external sensors
    for entity in entities do
        lvars dist = sim_distance(agent, entity);
        ;;; check if within some sensor range to speed things up...
        unless (entity == agent or
                dist > maxrange or
                sim_status(entity) == "dead") then
            lvars ex = sim_x(entity),
                  ey = sim_y(entity),
                  dir = sim_heading_from(ax, ay, ex, ey);
            if isfoodforagent(agent,entity) then
                if dist <= ingesting_rng then
                    if iswater(entity) then
                        conspair([ingest water],touchitems) -> touchitems;
                        entity -> wateritem(agent);
                    else
			;;; plants or other stuff that counts as food for
			;;; the agent
                        conspair([ingest food],touchitems) -> touchitems;
                        entity -> fooditem(agent);
                    endif;
                elseif dist <= smell_rng then
                    if iswater(entity) then
                        get_force(ax, ay, ex, ey, waterfactor);
                        + watery -> watery;
                        + waterx -> waterx;
                    else
                        get_force(ax, ay, ex, ey, foodfactor);
                        + foody -> foody;
                        + foodx -> foodx;
                    endif;
                endif;
                ;;; sense food and water that is within the vision range
                    if in_range(dist,vision_rng,ah,dir,60) then
                        if iswater(entity) then
                            conspair([water %dir% %dist% %min_dist(entity)%],fooditems)
                                -> fooditems;
                        else
                            conspair([food %dir% %dist% %min_dist(entity)%],fooditems)
                                -> fooditems;
                        endif;
                    endif;
             ;;; only obstacles and other agents
            elseunless isfood(entity) then
		;;; check if within collision range of 80 degress in front
		;;; of the agent
                if in_range(dist,immanent_collision_range(agent),ah,dir,40) then
                    if (isbasic_agent(entity)) then
                        if (datakey(entity) == datakey(agent)) then
                            true ->> sameagentcolliding -> agent.same_alarm;
                        else
                            true ->> othercolliding -> agent.other_alarm;
                        endif;
                        true ->> agentcolliding -> agent.agent_alarm;
                    else
                        true ->> obstcolliding -> agent.obst_alarm;
                    endif;
                endif;

		;;; ---------------- COLLISION CHECK --------------------------
		;;; TODO: when agents feed on each other, collision may not
		;;; be counted as such
		;;; TODO: the first part is for agents that have different
		;;; shapes, not yet used
		if shape(agent) then
		    ;;; two polys
		    if shape(entity) then
		    ;;; one poly one other
		    else
		    endif;
		else
		    ;;; agents vs. obstacles...
		    ;;; slot shape = {{-9 -9} {-9 9} {9 9} {9 -9}};

		    ;;; These formulas determine whether and where a circle
		    ;;; intersects a line.  Check all vectors.  If any
		    ;;; intersection exists on a line segment, there is a
		    ;;; collision.

		    ;;; dx = x2 - x1
		    ;;; dy = y2 - y1
		    ;;; dr = sqrt(dx^2 + dy^2)
		    ;;; D = (x1 * y2) - (x2 * y1)
		    ;;;	    (D * dy) +- sgn(dy) * dx * sqrt(r^2 * dr^2 - D^2)
		    ;;; x = -------------------------------------------------
		    ;;;                       dr^2
		    ;;;     (-D * dx) +- abs(dy) * sqrt(r^2 * dr^2 - D^2)
		    ;;; y = ---------------------------------------------
		    ;;;                       dr^2
		    ;;; sgn(x) = -1 for x < 0, 1 otherwise
		    ;;; delta = r^2 * dr^2 - D^2
		    ;;; if delta < 0, no intersection
		    ;;;	   delta = 0, tangent (one point)
		    ;;;    delta > 0, intersection (two points)

		    if shape(entity) then
			lvars ax = sim_x(agent),
			      ay = sim_y(agent),
			      ex = sim_x(entity),
			      ey = sim_y(entity),
			      crd = shape(entity),
			      len = length(crd),
			      collision = false,
			      r = min_dist(agent),
			      dx, dy, dr, D, delta, one, two, sgn, x, y;
			for one from 1 to len do
			    lvars x1, y1, x2, y2, plumin;
			    ;;; find next vertex
			    if one < len then one + 1 -> two;
			    else 1 -> two endif;
			    ;;; have to shift vertices relative to circle
			    ex + crd(one)(1) - ax -> x1;
			    ey + crd(one)(2) - ay -> y1;
			    ex + crd(two)(1) - ax -> x2;
			    ey + crd(two)(2) - ay -> y2;
			    x2 - x1 -> dx;
			    y2 - y1 -> dy;
			    sqrt(dx**2 + dy**2) -> dr;
			    ;;; compute sgn(dy)
			    if dy < 0 then -1 -> sgn;
			    else 1 -> sgn endif;
			    (x1 * y2) - (x2 * y1) -> D;
			    r**2 * dr**2 - D**2 -> delta;
			    ;;; Check for zero intercepts
			    nextif(delta < 0);
			    ;;; Vertical line; find y intercepts and check
			    ;;; to see if they're in between the endpoints
			    ;;; of the edges
			    if x1 = x2 then
				abs(dy) * sqrt(r**2 * dr**2 - D**2) -> plumin;
				(-D * dx + plumin) / dr**2 -> y;
				if (y >= y1 and y <= y2) or
				   (y >= y2 and y <= y1) then
				    true -> collision;
				    quitloop;
				endif;
				;;; Check for tangent
				nextif(delta <= 0);
				(-D * dx - plumin) / dr**2 -> y;
				if (y >= y1 and y <= y2) or
				   (y >= y2 and y <= y1) then
				    true -> collision;
				    quitloop;
				endif;
			    ;;; Non-vertical line; find x intercepts and
			    ;;; check to see if they're in between the
			    ;;; endpoints of the edges
			    else
				sgn * dx * sqrt(r**2 * dr**2 - D**2) -> plumin;
				(D * dy + plumin) / dr**2 -> x;
				if (x >= x1 and x <= x2) or
				   (x >= x2 and x <= x1) then
				    true -> collision;
				    quitloop;
				endif;
				;;; Check for tangent
				nextif(delta <= 0);
				(D * dy - plumin) / dr**2 -> x;
				if (x >= x1 and x <= x2) or
				   (x >= x2 and x <= x1) then
				    true -> collision;
				    quitloop;
				endif;
			    endif;
			endfor;
		    ;;; This formula determines how far it is from a point
		    ;;; to a line.  Use it to tell if the center of the
		    ;;; circle is inside the poly.  Check vectors
		    ;;; proceeding clockwise (or counterclockwise) around
		    ;;; the poly.  If all values are of the same sign, the
		    ;;; center is inside the poly and there is a collision.

		    ;;; (y2 - y1)*(x0 - x1) - (x2 - x1)*(y0 - y1)
		    ;;; -----------------------------------------
		    ;;;      sqrt((x2 - x1)^2 + (y2 - y1)^2)
		
			lvars side = 0;
			for one from 1 to len do
			    lvars x1, y1, x2, y2, dist;
			    quitif(collision);
			    ;;; find next vertex
			    if one < len then one + 1 -> two;
			    else 1 -> two endif;
			    ;;; find vertex coordinates
			    ex + crd(one)(1) -> x1;
			    ey + crd(one)(2) -> y1;
			    ex + crd(two)(1) -> x2;
			    ey + crd(two)(2) -> y2;
			    ;;; calculate distance from center to line
			    (y2 - y1)*(ax - x1) - (x2 - x1)*(ay - y1) -> dist;
			    dist / sqrt((x2 - x1)**2 + (y2 - y1)**2) -> dist;
			    ;;; Boom
			    if dist = 0 then
				true -> collision;
				quitloop;
			    ;;; First side checked
			    elseif side = 0 then
				if dist < 0 then
				    -1 -> side;
				else
				    1 -> side;
				endif;
			    elseif side = -1 then
				;;; Wrong side
				quitif(dist > 0);
			    else
				;;; Wrong side
				quitif(dist < 0);
			    endif;
			    ;;; If all sides checked without quitting loop,
			    ;;; then center is inside entity
			    if one = len then
				true -> collision;
				quitloop;
			    endif;
			endfor;
			if collision then
			    dies_basic_agent(agent,"collision");
			endif;
		    else
			if (dist < min_dist(entity) + min_dist(agent)) then
			    dies_basic_agent(agent,"collision");
			    if isbasic_agent(entity) then
				dies_basic_agent(entity,"collision");
			    endif;
			endif;
		    endif;
		endif;
		;;; ------------------------------------------------
                ;;; check if obstacle or agent is within sonar range
                if dist <= sonar_rng then
                    if (isbasic_agent(entity)) then
                        if (datakey(entity) == datakey(agent)) then
			    get_force(ax, ay, ex, ey, samefactor);
			    + samey -> samey;
			    + samex -> samex;
                        else
                            get_force(ax, ay, ex, ey, otherfactor);
                            + othery -> othery;
                            + otherx -> otherx;
                        endif;
                        get_force(ax, ay, ex, ey, agentfactor);
                        + agenty -> agenty;
			+ agentx -> agentx;
                    else
                        get_force(ax, ay, ex, ey, obstfactor);
	                + obsty -> obsty;
    	                + obstx -> obstx;
		    endif;
                endif;
                  if (dist <= touch_rng) then
                        [touch %dir%] :: touchitems -> touchitems;
                    endif;

                    if in_range(dist,vision_rng,ah,dir,60) then
                      if isobstacle(entity) then
                        conspair([obstacle %dir% %dist% %min_dist(entity)%],
                             visionitems) -> visionitems;
                      elseif (datakey(entity) == datakey(agent)) then
                        conspair([same %dir% %dist% %min_dist(entity)% %energy(entity)% %coolant(entity)%],
                             visionitems) -> visionitems;
                      else
                        conspair([other %dir% %dist% %min_dist(entity)% %energy(entity)% %coolant(entity)%],
                             visionitems) -> visionitems;
		              endif;
                    endif;
            endif;
        endunless;
    endfor;
    ;;; each vector represents a force in the respective sensory vector field
    consvector(agentx,agenty,2) -> agentvector(agent);
    consvector(samex,samey,2) -> samevector(agent);
    consvector(otherx,othery,2) -> othervector(agent);
    consvector(waterx,watery,2) -> watervector(agent);
    consvector(obstx,obsty,2)   -> obstvector(agent);
    consvector(foodx,foody,2)   -> foodvector(agent);
    [[foodvector %foodx, foody%]
     [watervector %waterx, watery%]
     [obstaclevector %obstx, obsty%]
     [agentvector %agentx, agenty%]
     [samevector %samex, samey%]
     [othervector %otherx, othery%]
     [agentcollision %agentcolliding%]
     [sameagentcollision %sameagentcolliding%]
     [othercollision %othercolliding%]
     [obstcollision %obstcolliding%]
     [collision %obstcolliding or agentcolliding%]]
     <> touchitems -> sensor_data;
	sensor_data <> fooditems <> visionitems <> snditems -> sensor_data;
    ;;; add proprioception
    [[energylevel %energy(agent)%]
     [waterlevel %coolant(agent)%]
     [heading %heading(agent)%]
     [speed %speed(agent)%]]
     <> sensor_data -> sensor_data;
enddefine;

;;; ------------------------ internal methods -----------------------------

;;; update agent body according to the move it made, no GUI
define :method update_body(agent:basic_agent);
    ;;; get the current speed
    lvars spd = speed(agent);

    ;;; food decreases quadratically, 1 unit for basic operation
    energy(agent) - spd*spd/2.0 - 1 -> energy(agent);

    ;;; water decreases quadratically, 1 unit for basic operation
    coolant(agent) - spd*spd/3.0 - 1 -> coolant(agent);

    ;;; update the agent age as well
    age(agent) + 1 -> age(agent);

    ;;; reduce procreation latency
    if procreation_latency(agent) > 0 then
        procreation_latency(agent) - 1 -> procreation_latency(agent);
    endif;

    ;;; check if the agent ran out of coolant or energy
    if (energy(agent) < 1) then
        dies_basic_agent(agent,"hunger");
    elseif (coolant(agent) < 1) then
        dies_basic_agent(agent,"thirst");
    endif;
	
    ;;; mark the agent as updated (so it will only be updated once per cycle
    true -> updated(agent);
enddefine;

;;; ------------------------ effector methods -----------------------------

;;; move action for agents
;;; if newx and newy are 0, then the agent moves in the direction determined
;;; by heading, otherwise the new heading is determined from newx and newy
define :method move_basic_agent(agent:basic_agent, newx, newy);
    lvars x = sim_x(agent),
          y = sim_y(agent),
        sp = speed(agent),
		k = class_name(datakey(agent)),
        new_heading = (if newx = 0 and newy = 0 then heading(agent) else
                          sim_heading_from(x, y, x + newx, y + newy) endif);

    lvars xnew, ynew;
    ;;; this uses the simworld clipping function in accordance with the world model
    update_position(x + sp * cos(new_heading),y + sp * sin(new_heading)) -> (xnew,ynew);
    ;;; update statistics
    agent_distance(k) + sp -> agent_distance(k);

    ;;; now set new coordinates and heading
    xnew -> sim_x(agent);
    ynew -> sim_y(agent);
    new_heading -> heading(agent);

    ;;; update bodily parameters that change because of the move
    update_body(agent);
enddefine;


;;; moves the agent into the new direction regardless of sensor fields
define :method reflex_basic_agent(agent:basic_agent);
    ;;; Move action for agents

    unless (sim_status(agent) == "paralyzed") then
        lvars s = speed(agent);
        reflexspeed(agent) -> speed(agent);
        ;;; move agent in the direction in heading set by the rule system
        move_basic_agent(agent,0,0);
        s -> speed(agent);
    endunless;
enddefine;

;;; combines sensors and maps sensor space onto motor space
define :method map_motor(agent:basic_agent);
    samevector(agent)(1) + obstvector(agent)(1)
    + foodvector(agent)(1) + watervector(agent)(1)
    + othervector(agent)(1)  -> movevector(agent)(1);

    samevector(agent)(2) + obstvector(agent)(2)
    + foodvector(agent)(2) + watervector(agent)(2)
    + othervector(agent)(2) -> movevector(agent)(2);
enddefine;

;;; moves the agent along the trajectories computes by various vectors...
define :method moving(agent:basic_agent);
	dlocal prb_database = sim_data(agent);
    unless prb_in_database([halting_for ==]) then
    	move_basic_agent(agent,movevector(agent)(1),movevector(agent)(2));
    endunless;
enddefine;

define :method eating(agent:basic_agent);
    lvars intake = maxfoodintake(agent),
       oldenergy = energy(agent),
		  k = kind(agent),
          entity = fooditem(agent);
	dlocal prb_database = sim_data(agent);

    if fooditem(agent) then
      if oldenergy >= maxenergy(agent) then
		;;; sated, stop eating
        prb_flush1([halting_for eating]);
      elseif (energy(entity) <= intake) then
        oldenergy + energy(entity) -> energy(agent);
        ;;; for statistics record the energy intake
        agent_energy(k) + energy(entity) -> agent_energy(k);
        dies_entity(entity);
      else
        oldenergy + intake -> energy(agent);
        ;;; for statistics record the energy intake
        agent_energy(k) + intake -> agent_energy(k);
        energy(entity) - intake -> energy(entity);
      endif;
    endif;
enddefine;


define :method drinking(agent:basic_agent);
    lvars intake = maxwaterintake(agent),
      oldcoolant = coolant(agent),
		  k = kind(agent),
          entity = wateritem(agent);
	dlocal prb_database = sim_data(agent);

  if wateritem(agent) then
    if oldcoolant >= maxcoolant(agent) then
		;;; sated, stop drinking
        prb_flush1([halting_for drinking]);
    elseif (energy(entity) <= intake) then
        oldcoolant + energy(entity) -> coolant(agent);
        ;;; for statistics record the energy intake
        agent_water(k) + energy(entity) -> agent_water(k);
        dies_entity(entity);
    else
        oldcoolant + intake -> coolant(agent);
        ;;; for statistics record the energy intake
        agent_water(k) + intake -> agent_water(k);
        energy(entity) - intake -> energy(entity);
    endif;
  endif;
enddefine;

;;; creates a new heading in the current direction +- range + oldheading
define :method newrandomheading(agent:basic_agent,range,dir);
    lvars newhead = dir + range - random(2*range) + heading(agent);

    if newhead > 360 then
        newhead - 360 -> heading(agent)
    else
        newhead -> heading(agent)
    endif;
enddefine;

;;; creates a new heading in the current direction +- range + oldheading
define :method turning(agent:basic_agent,dir);
    lvars newhead = dir + heading(agent);

    if newhead > 360 then
        newhead - 360 -> heading(agent)
    else
        newhead -> heading(agent)
    endif;
enddefine;


;;; increase the speed
define :method speeding_up(agent:basic_agent);
    if speed(agent) < maxspeed(agent) then
       speed(agent) + 1 -> speed(agent)
    endif;
enddefine;

;;; decrease the speed
define :method slowing_down(agent:basic_agent);
    if speed(agent) > 0 then
       speed(agent) - 1 -> speed(agent)
    endif;
enddefine;

;;; halt
define :method halting_for(agent:basic_agent, why);
    dlocal prb_database = sim_data(agent);

    ;;;npr('--Halting for ' >< why);
    prb_add([halting_for %why%]);
    0 -> speed(agent);
enddefine;
