/*********************************************************************

SIMWORLD
Version 4.0 Beta

Basic Agents GUI

Copyright (C) 2002 Matthias Scheutz (mscheutz@nd.edu)
Artificial Intelligence and Robotics Laboratory
Department of Computer Science and Engineering 
University of Notre Dame, USA
http://www.nd.edu/~mscheutz/ 
http://www.nd.edu/~airolab/ 
All rights reserved.

This program is licensed and distributed under the terms stated
in the accompanying COPYRIGHT file.

Last modified: 11-27-02

*********************************************************************/

;;; note that this preserves the hierarchy among the nonGUI classes, but
;;; does not establish a similar hierarchy among GUI classes, which is not
;;; needed anyway!

;;; this is the superclass of all GUIs
define :class GUI; is sim_multiwin_mobile;
    slot sim_windows == false;            ;;; the window(s) where the agent should be displayed
    slot sim_x == false;
    slot sim_y == false;
enddefine;

define :class thingGUI; is GUI, thing; enddefine;
define :class obstacleGUI; is GUI, obstacle; enddefine;
define :class moving_obstacleGUI; is moving_obstacle, obstacleGUI; enddefine;
define :mixin foodGUI; is GUI, food; enddefine;
define :class plantGUI; is plant, foodGUI; enddefine;
define :class waterGUI; is water, foodGUI; enddefine;


;;; this defines the basic GUI class for agents, which every GUI agent inherits
;;; can be used to check whether an agent is an instance of a GUI class
define :class basic_agentGUI; is GUI, basic_agent,
                           rc_rotatable_picsonly;

    ;;; this is used for drawing various sensor ranges
    slot drawvisionrange = SHOWVISION;    ;;; indicates whether vision range should be drawn
    slot drawsonarrange = SHOWSONAR;      ;;; indicates whether sonar range should be drawn
    slot drawsmellrange = SHOWSMELL;      ;;; indicates whether smell range should be drawn
    slot drawpath = SHOWPATH;             ;;; indicates whether the path should be drawn
    slot drawstate = SHOWSTATE;           ;;; indicates whether the state of the agent should be shown

    ;;; this is for drawing vectors...
    slot oldfoodvector == {%0,0%};
    slot oldwatervector == {%0,0%};
    slot oldothervector == {%0,0%};
    slot oldsamevector == {%0,0%};
    slot oldagentvector == {%0,0%};
    slot oldobstvector == {%0,0%};

    ;;; this will display the database right after sensing if the corresponding ruleset is included
    slot printdatabase == true;

    ;;; for drawing the force vectors
    slot drawvectors = SHOWVECTORS;       ;;; draws the respective vectors
    slot mainpanel == false;          ;;; will hold the main panel for this agent
enddefine;

define :method update_body(agent:basic_agentGUI);
    ;;; call the update_body function for basic_agent
    call_next_method(agent);

    ;;; undraw old picture
    rc_draw_oldpic(agent);

    if drawstate(agent) then
            [FONT '8x13' {10 0 %sim_name(agent) >< ' ' >< energy(agent) >< ','
                          >< coolant(agent) >< ',' >< speed(agent) %}] -> rc_pic_strings(agent);
    else
        [FONT '8x13' {10 0 %sim_name(agent) >< ''%}] -> rc_pic_strings(agent);
    endif;
    ;;; account for adjustments to the heading
    heading(agent) -> rc_axis(agent);
    rc_draw_linepic(agent);
enddefine;

;;; the moving function for the GUI agents
define :method move_basic_agent(agent:basic_agentGUI,newx,newy);
    ;;; get old agent coordinates
    lvars x = sim_x(agent),y = sim_y(agent);

    ;;; call the move_basic_agent function
    call_next_method(agent,newx,newy);

    ;;; save current foreground color and get new agent coordinates
    lvars w = if islist(sim_windows(agent)) then hd(sim_windows(agent)) else sim_windows(agent) endif,
	 fg = rc_foreground(rc_window),
         nx = sim_x(agent),ny = sim_y(agent);

    ;;; This will be done only to the first window in the list assuming that there is only
    ;;; one "detail" window...

	hd(sim_windowlist) -> rc_current_window_object;
    ;;; DO I NEED TO MAKE THE WINDOW CURRENT, i.e., rc_current_window_object ???
    if drawpath(agent) then
        bodycolor(agent) -> rc_foreground(rc_window);
        rc_drawline(x,y,nx,ny);
        fg -> rc_foreground(rc_window);
    endif;
    ;;; undraw old and draw new vectors
    if drawvectors(agent) then
        dlocal rc_linefunction = Glinefunction; ;;; rc_window = sim_window;
        'red' -> rc_foreground(rc_window);
        rc_drawline(x,y,x-oldobstvector(agent)(1)*100,y-oldobstvector(agent)(2)*100);
        rc_drawline(nx,ny,nx-obstvector(agent)(1)*100,ny-obstvector(agent)(2)*100);
        'brown' -> rc_foreground(rc_window);
        rc_drawline(x,y,x-oldagentvector(agent)(1)*100,y-oldagentvector(agent)(2)*100);
        rc_drawline(nx,ny,nx-agentvector(agent)(1)*100,ny-agentvector(agent)(2)*100);
        'white' -> rc_foreground(rc_window);
        rc_drawline(x,y,x-oldothervector(agent)(1)*100,y-oldothervector(agent)(2)*100);
        rc_drawline(nx,ny,nx-othervector(agent)(1)*100,ny-othervector(agent)(2)*100);
        'green' -> rc_foreground(rc_window);
        rc_drawline(x,y,x+oldfoodvector(agent)(1)*100,y+oldfoodvector(agent)(2)*100);
        rc_drawline(nx,ny,nx+foodvector(agent)(1)*100,ny+foodvector(agent)(2)*100);
        'lightblue' -> rc_foreground(rc_window);
        rc_drawline(x,y,x+oldwatervector(agent)(1)*100,y+oldwatervector(agent)(2)*100);
        rc_drawline(nx,ny,nx+watervector(agent)(1)*100,ny+watervector(agent)(2)*100);
        fg -> rc_foreground(rc_window);
    endif;

	;;; keep track of old vectors...
    agentvector(agent) -> oldagentvector(agent);
    samevector(agent) -> oldsamevector(agent);
    othervector(agent) -> oldothervector(agent);
    obstvector(agent)  -> oldobstvector(agent);
    foodvector(agent)  -> oldfoodvector(agent);
    watervector(agent) -> oldwatervector(agent);
enddefine;



;;; this will set the new linepics
define :method update_ranges(agent:basic_agentGUI);
    rc_draw_oldpic(agent);
    [
     [WIDTH 2 COLOUR %bodycolor(agent)%
      [CIRCLE {0 0 %min_dist(agent)%}]
      [COLOUR 'black' [SQUARE {12 0 2}]]]
     ] -> rc_pic_lines(agent);
	hd(sim_windowlist) -> rc_current_window_object;
    if (drawvisionrange(agent)) then
        conspair([WIDTH 1 COLOUR 'black' STYLE ^LineOnOffDash
                  [ANGLE 60 {0 0} {%vision_range(agent)% 0}]
                  [ANGLE 300 {0 0} {%vision_range(agent)% 0}]
                  [CIRCLE {0 0 %vision_range(agent)%}]],
        rc_pic_lines(agent)) -> rc_pic_lines(agent);
    endif;
    if (drawsonarrange(agent)) then
        conspair([WIDTH 1 COLOUR 'red' STYLE ^LineOnOffDash
                  [CIRCLE {0 0 %sonar_range(agent)%}]],
        rc_pic_lines(agent)) -> rc_pic_lines(agent);
    endif;
    if (drawsmellrange(agent)) then
        conspair([WIDTH 1 COLOUR 'lightgreen' STYLE ^LineOnOffDash
                  [CIRCLE {0 0 %smell_range(agent)%}]],
        rc_pic_lines(agent)) -> rc_pic_lines(agent);
    endif;
    rc_draw_linepic(agent);
enddefine;

;;; this will update the vision range whenever it is changed
define :wrapper updaterof drawvisionrange(val, agent:basic_agentGUI, upd_p);
    upd_p(val,agent);
    update_ranges(agent);
enddefine;

;;; this will update the sonar range whenever it is changed
define :wrapper updaterof drawsonarrange(val,agent:basic_agentGUI, upd_p);
    upd_p(val,agent);
    update_ranges(agent);
enddefine;

;;; this will update the smell range whenever it is changed
define :wrapper updaterof drawsmellrange(val,agent:basic_agentGUI, upd_p);
    upd_p(val,agent);
    update_ranges(agent);
enddefine;

;;; this will update the draw vectors and make them draw right...
define :wrapper updaterof drawvectors(val,agent:basic_agentGUI, upd_p);
	hd(sim_windowlist) -> rc_current_window_object;
    dlocal rc_linefunction = Glinefunction;
    lvars x = sim_x(agent),y = sim_y(agent), fg = rc_foreground(rc_window);

    ;;; if in GUI mode just draw over the current ones...
    ;;;     dlocal rc_linefunction = Glinefunction; ;;; rc_window = sim_window;
    'red' -> rc_foreground(rc_window);
    rc_drawline(x,y,x-obstvector(agent)(1)*100,y-obstvector(agent)(2)*100);
    'brown' -> rc_foreground(rc_window);
    rc_drawline(x,y,x-agentvector(agent)(1)*100,y-agentvector(agent)(2)*100);
    'white' -> rc_foreground(rc_window);
    rc_drawline(x,y,x-othervector(agent)(1)*100,y-othervector(agent)(2)*100);
    'green' -> rc_foreground(rc_window);
    rc_drawline(x,y,x+foodvector(agent)(1)*100,y+foodvector(agent)(2)*100);
    'lightblue' -> rc_foreground(rc_window);
    rc_drawline(x,y,x+watervector(agent)(1)*100,y+watervector(agent)(2)*100);
    fg -> rc_foreground(rc_window);
    ;;; and set the value
    upd_p(val,agent);
enddefine;

;;; --------------------------------------------------------------------
;;; sets up various graphical representations of agent parameters
;;; adds the agent to a container or a list of containers
define :method show(agent:basic_agentGUI);
    ;;; set the heading...
	;;; heading(agent) -> rc_axis(agent);
    rc_set_axis(agent,heading(agent),true);
    [FONT '8x13' {10 0 %sim_name(agent) >< nullstring%}]
        -> rc_pic_strings(agent);
    [
        [WIDTH 2 COLOUR %bodycolor(agent)%
            [CIRCLE {0 0 %min_dist(agent)%}]
            [COLOUR 'black' [SQUARE {12 0 2}]]]
    ] -> rc_pic_lines(agent);
    ;;; draw various sensor ranges--assumes that they stay the same through
    ;;; the simulation
    if (drawvisionrange(agent)) then
        conspair([WIDTH 1 COLOUR 'black' STYLE ^LineOnOffDash
                    [ANGLE 60 {0 0} {%vision_range(agent)% 0}]
                    [ANGLE 300 {0 0} {%vision_range(agent)% 0}]
                    [CIRCLE {0 0 %vision_range(agent)%}]],
                 rc_pic_lines(agent)) -> rc_pic_lines(agent);
    endif;
    if (drawsonarrange(agent)) then
        conspair([WIDTH 1 COLOUR 'red' STYLE ^LineOnOffDash
                    [CIRCLE {0 0 %sonar_range(agent)%}]],
                 rc_pic_lines(agent)) -> rc_pic_lines(agent);
    endif;
    if (drawsmellrange(agent)) then
        conspair([WIDTH 1 COLOUR 'lightgreen' STYLE ^LineOnOffDash
                    [CIRCLE {0 0 %smell_range(agent)%}]],
                 rc_pic_lines(agent)) -> rc_pic_lines(agent);
    endif;

    ;;;NOT NEEDED store a copy of the linepic in here
    ;;;    if drawvectors(agent) then
    ;;;   copy(rc_pic_lines(agent))->drawvectors(agent);
    ;;;endif;
    if islist(sim_windows(agent)) then
        rc_add_containers(agent,sim_windows(agent));
    else
        rc_add_container(agent,sim_windows(agent));
    endif;

    ;;; if in interactive mode, create identifiers for easy reference to agents
    ;;; create an identifier for the name
    ident_declare(sim_name(agent), 0, 0);
	;;; return a reference to the agent
    agent -> valof(sim_name(agent));
enddefine;

define :method show(agent:obstacleGUI);
      [COLOUR %bodycolor(agent)% WIDTH 1 [CLOSED %explode(shape(agent))%] ]
        -> rc_pic_lines(agent);

    if islist(sim_windows(agent)) then
        rc_add_containers(agent,sim_windows(agent));
    else
        rc_add_container(agent,sim_windows(agent));
    endif;

    ;;; if in interactive mode, create identifiers for easy reference to agents
    ;;; create an identifier for the name
    ident_declare(sim_name(agent), 0, 0);
    agent -> valof(sim_name(agent));
enddefine;

define :method show(obj:foodGUI);
    [COLOUR %bodycolor(obj)% WIDTH 1 [CIRCLE {0 0 %min_dist(obj)%}] ]
        -> rc_pic_lines(obj);
    if islist(sim_windows(obj)) then
        rc_add_containers(obj,sim_windows(obj));
    else
        rc_add_container(obj,sim_windows(obj));
    endif;
    ;;; if in interactive mode, create identifiers for easy reference to agents
    ;;; create an identifier for the name
    ident_declare(sim_name(obj), 0, 0);
    obj -> valof(sim_name(obj));
enddefine;


;;;------------------------- various methods for GUI mode----------------------

;;; for moving obstacles, override sim_run_agent with simpler version
define :method sim_run_agent(obst:moving_obstacleGUI, objects);
    rc_draw_oldpic(obst);

    ;;; call the sim_run_agent function for moving_obstacle
    call_next_method(obst, objects);

    rc_draw_linepic(obst);
enddefine;

;;; this removes an arbitrary entity from the simulation
define :method dies_entity(entity:GUI);

    call_next_method(entity);
    rc_draw_oldpic(entity);

    if islist(sim_windows(entity)) then
	lvars w;
	for w in sim_windows(entity) do
	    rc_remove_pic_from_window(entity, w);
	endfor;
    else
	rc_remove_pic_from_window(entity, sim_windows(entity));
    endif;	
    [] -> rc_pic_containers(entity);
enddefine;

define :method dies_basic_agent(agent:basic_agentGUI,cause);

	call_next_method(agent,cause);
	;;; if vectors are shown, undraw them
	if drawvectors(agent) then
		lvars x = sim_x(agent), y = sim_y(agent), fg = rc_foreground(rc_window);
        dlocal rc_linefunction = Glinefunction; ;;; rc_window = sim_window;
        'red' -> rc_foreground(rc_window);
        rc_drawline(x,y,x-obstvector(agent)(1)*100,y-obstvector(agent)(2)*100);
        'brown' -> rc_foreground(rc_window);
        rc_drawline(x,y,x-agentvector(agent)(1)*100,y-agentvector(agent)(2)*100);
        'white' -> rc_foreground(rc_window);
        rc_drawline(x,y,x-othervector(agent)(1)*100,y-othervector(agent)(2)*100);
        'green' -> rc_foreground(rc_window);
        rc_drawline(x,y,x+foodvector(agent)(1)*100,y+foodvector(agent)(2)*100);
        'lightblue' -> rc_foreground(rc_window);
        rc_drawline(x,y,x+watervector(agent)(1)*100,y+watervector(agent)(2)*100);
        fg -> rc_foreground(rc_window);
    endif;

    rc_draw_oldpic(agent);

    if islist(sim_windows(agent)) then
	lvars w;
	for w in sim_windows(agent) do
	    rc_remove_pic_from_window(agent, w);
	endfor;
    else
	rc_remove_pic_from_window(agent, sim_windows(agent));
    endif;	
    [] -> rc_pic_containers(agent);
enddefine;


;;;------------------------------------------------------------------------------

;;; this function sets the windows slot of a GUI object to the current global
;;; sim_windowlist and then calls show on the object, returning the object
define :method initialize(obj:GUI,parents);
    call_next_method(obj,parents);
    ;;; need to use valof to make it revaluate the window list every time
    valof("sim_windowlist")->sim_windows(obj);
    show(obj);
enddefine;	
