/*********************************************************************

SIMWORLD                                                */ global vars
Version = '4.0 Beta   '                                            ;/*

Simulator

Copyright (C) 2002 Matthias Scheutz (mscheutz@nd.edu)
Artificial Intelligence and Robotics Laboratory
Department of Computer Science and Engineering
University of Notre Dame, USA
http://www.nd.edu/~mscheutz/
http://www.nd.edu/~airolab/
All rights reserved.

This program is licensed and distributed under the terms stated
in the accompanying COPYRIGHT file.

Last modified: 11-27-02

*********************************************************************/
;;; this may have to be adjusted
constant path_to_simworld_files = sys_fname_path(popfilename);

;;; Increase popmemlim to reduce garbage collection time.
max(popmemlim, 32000000) -> popmemlim;

;;; generate a random initial number--this can be set to any previously
;;; determined seed to restart the simulation in the same state
random(100000) -> ranseed;

;;; ignore type checks for procedure variables
true -> popdefineprocedure;

;;; ------------------- External libraries --------------------------
uses newkit;
uses objectclass;
uses simlib;
uses sim_agent;

;;; --------------- GLOBAL SIMULATION PARAMETERS -------------------
;;; define the server host and its port as default
;;; start in client mode, and assume that server and port are
;;; set correctly by the calling program if different from the
;;; default values below

global vars
    ;;; GLOBALS for STATISTICS
    ;;; the frequency with which the global simulation data is sampled
    RECFREQ = 200,

    ;;; to record the status of objects at frequency RECFREQ	
    alive_freq = newproperty([], 16, 0, "perm"),

    ;;; for global simulation statistics used by the server, this is
    ;;; recorded for each agent kind
    ;;; these are absolut numbers for each agent kind
    agent_alive = newproperty([], 16, 0.0, "perm"),
    agent_hunger = newproperty([], 16, 0.0, "perm"),
    agent_thirst = newproperty([], 16, 0.0, "perm"),
    agent_crash = newproperty([], 16, 0.0, "perm"),
    agent_eaten = newproperty([], 16, 0.0, "perm"),

    ;;; these are averages over the total number of agents of a kind in one
    ;;; simulation run
    agent_age = newproperty([], 16, 0.0, "perm"),
    agent_energy = newproperty([], 16, 0.0, "perm"),
    agent_water = newproperty([], 16, 0.0, "perm"),
    agent_distance = newproperty([], 16, 0.0, "perm"),
    agent_offspring = newproperty([], 16, 0.0, "perm"),

    ;;; these are compound efficiency measures that are computed at the end
    ;;; of the simulation only
    agent_eef = newproperty([], 16, 0.0, "perm"),
    agent_wef = newproperty([], 16, 0.0, "perm"),

    ;;; this is for individual recordings, needs to be false originally!
    RECINDSLOTS = newproperty([],10000,false,"perm"),
    RECIND = true,   ;;; to turn recording on

    ;;; now save all the stats in a single property
    allstats = newproperty([[agents ^RECINDSLOTS]
			    [sampling ^alive_freq]
                            [alive ^agent_alive]
                            [hunger ^agent_hunger]
                            [thirst ^agent_thirst]
                            [crash ^agent_crash]
                            [eaten ^agent_eaten]
                            [age ^agent_age]
                            [energy ^agent_energy]
                            [water ^agent_water]
                            [distance ^agent_distance]
                            [offspring ^agent_offspring]
                            [EEF ^agent_eef]
                            [WEF ^agent_wef]],14,false,"perm"),

    ;;; default offspring parameter
    OMEGA = 250,     ;;; the procreation age
    FERT = false,    ;;; the fertility period (from OMEGA to OMEGA+FERT)
    PROCINT = 0,    ;;; the latency between kids

    ;;; this turns mutation on or off
    MUTATION = true,

    ;;; sets a random seed and stores it
    originalseed = ranseed,

    ;;; for saving the sim state at periodic interval
    checkINTERVAL = false,
    useTIMER = false,
    imagename = 'myimage.psv',

    ;;; GLOBALS related to the sim_ libraries
    ;;; Maximum number of decimal places shown in trace printing
    sim_pr_places = 2,

    ;;; individual variables for controlling tracing and pausing
    /* See
        HELP POPRULEBASE/prb_sayif_trace
        This is a user definable list of words which are associated
        with SAYIF poprulebase actions.
    */
    prb_sayif_trace = [],
    sim_alltracevars =
        [
            ;;; Poprulebase variables (See HELP POPRULEBASE)
            prb_walk prb_chatty prb_show_conditions

            ;;; switch garbage collection tracing on or off
            popgctrace],
    ;;; set this false to disable PAUSE actions
    prb_pausing = false,

    ;;; the list and types of all agents created before and after a run of the
    ;;; simulation
    alive_agents = [],
    KEEPAGENTS = false,  ;;; set this to true to keep a list of dead agents...
    dead_agents = [],
    agenttypes = [],

    ;;; this is for the creation and initalization of entities, needs to be
    ;;; set to false by default!
    constructors = newproperty([], 16, false, "perm"),

    ;;; the dimensions of the environment
    world_width = 800,
    world_height = 800,
    world_lim_x = world_width/2,
    world_lim_y = world_height/2,

    ;;; the standard clipping procedure, can be redefined for wrapping
    ;;; or for open environments
    update_position = false,

    ;;; this is for reporting unknown keywords in the simulation parameters
    ;;; by default mishap on unknown keywords
    ignoreUNKNOWNKEYWORD = false,

    ;;; use basic agents
    useAGENTS = 'basic',
;

;;; for debugging: print out lots of comments
lconstant verbose = false;

;;; ----------  Utility functions -----------------------------------------

;;; generate random location, just inside the world limits limits
define random_xy() -> (x,y);
    lvars w = world_width*0.45, h = world_height*0.45;
    random(2*w) - w -> x;
    random(2*h) - h -> y;
enddefine;

;;; random value within (xo,yo) and (xo+w,yo+h) without clipping to window size
define randomrectangle(xo,yo,w,h) -> (x,y);
    xo + random(w) - w -> x;
    yo + random(h) - h -> y;
enddefine;

;;; generate random location on a cirle with radius r around x, y
define random_close_to(x,y,r) -> (x,y);
    lvars phi = random(360);
    r * cos(phi) + x -> x;
    r * sin(phi) + y -> y;
enddefine;

;;; generate random location within a cirle with radius r around x, y
define random_within(x,y,r) -> (x,y);
    lvars phi = random(360), l = random(r);
    l * cos(phi) + x -> x;
    l * sin(phi) + y -> y;
enddefine;

;;; various clipping functions
define to_worldx(x);
    if x > world_lim_x then
        world_lim_x
    elseif x < -world_lim_x then
        -world_lim_x
    else
	x
    endif;
enddefine;

define to_worldy(y);
    if y > world_lim_y then
        world_lim_y
    elseif y < -world_lim_y then
        -world_lim_y
    else
	y
    endif;
enddefine;

define wrap_aroundx(x);
    if x > world_lim_x then
        world_lim_x - (x - world_lim_x)
    elseif x < -world_lim_x then
        world_lim_x + (x + world_lim_x)
    else
	x
    endif;
enddefine;

define wrap_aroundy(y);
    if y > world_lim_y then
        world_lim_y - (y - world_lim_y)
    elseif y < -world_lim_y then
        world_lim_y + (y + world_lim_y)
    else
	y
    endif;
enddefine;

;;; set the standard clipping procedure for rectangular worlds
procedure(x,y)->(xn,yn);to_worldx(x)->xn;to_worldy(y)->yn;endprocedure -> update_position;

;;; this is used to get a word that names the class of an object
define kind = datakey <> class_name enddefine;

;;; ----------------- Load the other simworld files -----------------------
;;; this loads all agent types, will also load the planner
if sys_file_exists(path_to_simworld_files >< useAGENTS >< 'agents.p') then
    compile(path_to_simworld_files dir_>< useAGENTS >< 'agents.p');
else
    mishap('Specified agents file does not exist');
endif;
;;; this loads all rules
if sys_file_exists(path_to_simworld_files >< useAGENTS >< 'rules.p') then
    compile(path_to_simworld_files dir_>< useAGENTS >< 'rules.p');
else
    mishap('Specified rules file does not exist');
endif;
;;; this loads the rulesystems
if sys_file_exists(path_to_simworld_files >< useAGENTS >< 'rulesystems.p') then
    compile(path_to_simworld_files dir_>< useAGENTS >< 'rulesystems.p');
else
    mishap('Specified rulesystems file does not exist');
endif;
;;; this will be overwritten during setup to perform whatever action is needed
;;; each cycle, e.g., food and water can be added to the simulation
define update_resources(sim_objects,sim_cycle_number);
enddefine;

;;;------------ redefined SIMAGENT methods --------------------------------------
;;; this one is called regularly by sim_scheduler at the beginning to create objects
define vars procedure sim_setup_scheduler(sim_objects, sim_cycle_number);
    lvars aft = newproperty([],20, 0, "perm"),
     	  entity;

    update_resources(sim_objects,sim_cycle_number);
    for entity in sim_objects do
	;;; get the current kind
	lvars knd = kind(entity),
	      slotname = if gender(entity) then
		            consword(knd >< '_' >< 'gender(entity)')
	                 else
                            knd
			 endif;
        ;;; count for statistics
        aft(slotname) + 1 -> aft(slotname);

	;;; check if agent is ready to procreate
	if readytoprocreate(entity) then
	    ;;; create a new agent, initialize and possibly mutate it

	    lvars newagent = offspring(entity);
            if MUTATION then
		mutate(newagent);
	    endif;

	    ;;; set the procreation latency for the parent
            PROCINT -> procreation_latency(entity);
	
	    ;;; count the offspring
	    agent_offspring(knd) + 1 -> agent_offspring(knd);
	
	    ;;; mark the creation cycle in agent
	    sim_cycle_number -> createdat(newagent);

            ;;; add the agent to the simulation
            conspair(newagent,sim_object_add_list) -> sim_object_add_list;
            if verbose then
                npr(entity >< ' has offspring ' >< newagent);
                npr('Remaining energy:' >< energy(entity));
            endif;
        endif;
    endfor;

    ;;; TODO: fix this for GUI mode (it will not produce the right stats as it
    ;;; looks for the GUI agent types collect general information about the
    ;;; state of the simulation
    if sim_cycle_number rem RECFREQ = 0 then
        lvars item;
        for item in agenttypes do
            conspair(aft(item),alive_freq(item))->alive_freq(item);
        endfor;
    endif;
enddefine;

;;; this checks whether the state of the simulation should be saved regularly
define save_simstate(cycle);
    if useTIMER then
    	false -> pop_asts_enabled;
    endif;
    ;;; Suppress warning messages
    dlocal cucharerr = discout('/dev/null');
    if syssave(imagename) then
	if verbose then npr('Continuing simulation from cycle: ' >< cycle); endif;
    else
   	if verbose then npr('Saved simulation after ' >< cycle >< ' cycles'); endif;
    endif;
    if useTIMER then
    	true -> pop_asts_enabled;
    endif;
enddefine;

;;; this needs to run right after all actions of the objects are executed
define finish_up_cycle(objects, cycle);
    lvars entity;

    ;;; update agents again
    for entity in sim_objects do
        unless updated(entity) then
            update_body(entity);
	    false -> updated(entity);
        endunless;	
	if RECIND then
	    record(entity,cycle);
	endif
    endfor;
    ;;; check if we should perform checks, and if so whether we use cycles
    ;;; will save the sim state at period intervals if checkINTERVAL is a number
    if (not(useTIMER) and isnumber(checkINTERVAL) and (sim_cycle_number rem checkINTERVAL = 0)) then
        save_simstate(cycle);
    endif;
enddefine;

;;; this runs right after all actions of the objects are executed
;;; this will be overwritten by the GUI version, so make it call
;;; another method instead
define sim_scheduler_pausing_trace(objects, cycle);
    finish_up_cycle(objects,cycle);
enddefine;

;;; this one is called after each cycle
define vars procedure sim_post_cycle_actions(sim_objects, sim_cycle_number);
    ;;; CHECK FOR RUNTIME CONDITION...
    ;;;    true -> sim_stopping_scheduler;
enddefine;

;;; runs at the end of the initialization process
define finish_setup(alive_agents);
    ;;; Run at the end of setup_simulation
enddefine;


;;;--------- override various SIMAGENT tracing functions --------------

define :method sim_agent_running_trace(agent:sim_agent);
enddefine;

define :method sim_agent_messages_out_trace(agent:sim_agent);
enddefine;

define :method sim_agent_messages_in_trace(agent:sim_agent);
enddefine;

define :method sim_agent_actions_out_trace(object:sim_object);
enddefine;

define :method sim_agent_actions_out_trace(agent:sim_agent);
enddefine;

define :method sim_agent_action_trace(agent:sim_agent);
enddefine;

define :method sim_agent_rulefamily_trace(object:sim_object, rulefamily);
enddefine;

define :method sim_agent_rulefamily_trace(agent:sim_agent, rulefamily);
enddefine;

define :method sim_agent_endrun_trace(agent:sim_object);
enddefine;

define sim_scheduler_finished(objects, cycle);
enddefine;
;;;---------------------------------------------------------------------

;;; computes some more stats and prints out who is left after a whole simulation, in HTML or ASCII
define printagentstats(agents,which,format);
    lvars a = 0, c = 0, d = 0, r = 0,
          entity, mat = [], coef = [],
	  agent_sten = newproperty([], 16, 0.0, "perm"),
	  agent_stwa = newproperty([], 16, 0.0, "perm"),
	  agent_pren = newproperty([], 16, 0.0, "perm"),
	  agent_prwa = newproperty([], 16, 0.0, "perm");

    for entity in agents do
        if isbasic_agent(entity) then
	    ;;; this may write the same information many times, that's ok, just redundant
	    startupenergy(entity) -> agent_sten(kind(entity));
	    startupwater(entity) -> agent_stwa(kind(entity));
	    startupenergy(entity) -> agent_pren(kind(entity));
	    startupwater(entity) -> agent_prwa(kind(entity));
            agent_alive(kind(entity)) + 1 -> agent_alive(kind(entity));
	    agent_age(kind(entity)) + age(entity) -> agent_age(kind(entity));
        endif;
    endfor;

    ;;; print statistics as plain text or as HTML (using tables)
    lvars ag;
    if format == "HTML" then
        printf('<table>\n');
        printf('<tr><td>PREY</td><td>total</td><td>alive</td><td>crashed</td>'
	    >< '<td>eaten</td><td>hunger</td><td>thirst</td><td>av.age</td>'
	    >< '<td>av.eng</td><td>av.wat</td><td>av.dis</td><td>av.off</td>'
	    >< '<td>eff</td><td>wef</td></tr>\n');
    elseif format == "ASCII" then
	printf('PREY        total alive crashed eaten hunger thirst av.age'
	    >< 'av.energy av.water av.dist av.offspr eef. wef.\n');
    endif;
    for ag in which do
        lvars tot = agent_eaten(ag) +  agent_crash(ag) +
             agent_hunger(ag)+ agent_thirst(ag) + agent_alive(ag);
	npr('ag: ' >< ag >< ', tot = ' >< tot);
        if tot /= 0 then
	    npr('tot = ' >< tot);
	    lvars av_off = agent_offspring(ag)/tot,
		 av_dis = agent_distance(ag)/tot,
		 av_eng = agent_energy(ag)/tot,
 		 av_wat = agent_water(ag)/tot,
		 av_age = agent_age(ag)/tot,
		 ;;; get the energy and water expenditure per distance unit
		 ;;; travelled...
		 ;;; Need to exclude computation cost from these
		 ;;; calculations, hence the "- av_age" part.  In the
		 ;;; future, this will need to be more sophisticated.
		 av_eef = (agent_sten(ag) + av_eng - av_off * agent_pren(ag)
		    - av_age) / av_dis,
		 av_wef = (agent_stwa(ag) + av_wat - av_off * agent_prwa(ag)
		    - av_age) / av_dis;
	    if format == "HTML" then
		printf('<tr><td>%p:</td><td>%p</td><td>%p</td><td>%p</td>'
		>< '<td>%p</td><td>%p</td><td>%p</td><td>%p</td>'
		>< '<td>%p</td><td>%p</td><td>%p</td><td>%p</td>'
		>< '<td>%p</td><td>%p</td></tr>\n',
                [%ag,tot,agent_alive(ag),agent_crash(ag),agent_eaten(ag),
                agent_hunger(ag),agent_thirst(ag), av_age, av_eng, av_wat,
		av_dis, av_off, av_eef, av_wef%]);
	    elseif format == "ASCII" then
                printf('%p: %p  %p  %p  %p  %p  %p  %p  %p  %p  %p  %p  %p'
		>< '%p\n',
                [%ag,tot,agent_alive(ag),agent_crash(ag),agent_eaten(ag),
                agent_hunger(ag),agent_thirst(ag), av_age, av_eng, av_wat,
		av_dis, av_off, av_eef, av_wef%]);
	    endif;
	    ;;; save the averages in the respective properties, so they
	    ;;; won't have to be computed in the server
	    ;;; the original figures can be easily retrieved by multiplying
	    ;;; the values by total number of agents
	    av_off -> agent_offspring(ag);
	    av_dis -> agent_distance(ag);
	    av_eng -> agent_energy(ag);
 	    av_wat -> agent_water(ag);
	    av_age -> agent_age(ag);
	    av_eef -> agent_eef(ag);
	    av_wef -> agent_wef(ag);
        endif;
    endfor;
    if format == "HTML" then
	printf('</table>\n');
    endif;
enddefine;

;;; FIXME: needs to be fixed to print the list nicely
define printlist(params,printmode);
     if printmode == "HTML" then
	 pr('[ ');
     elseif printmode == "ASCII" then
	 pr('[');
     endif;
	 applist(params,procedure(e);if ispair(e) then printlist(e,printmode) else pr(e);pr(' '); endif; endprocedure);
     if printmode == "HTML" then
	 pr(']');	
     elseif printmode == "ASCII" then
	 pr(']');
     endif;
enddefine;

;;;------------------------ parsing setup file ------------------------------------------

;;; checks whether each s in sl is a slot of the class instantiated by ag
;;; s can also be just a single slot
define check_slots(sl,ag)->ret;
    lvars s, ret = true;

    unless islist(sl) then
        [%sl%] -> sl;
    endunless;

    for s in sl do
        unless isword(s) and isdeclared(s) and member(valof(s),class_slots(datakey(ag))) then
            npr('Invalid slot: ' >< s >< ' for agent ' >< sim_name(ag));
            false -> ret;
        endunless;
    endfor;
enddefine;

define check_initlist(il,ag);
    lvars s, csl = class_slots(valof(consword(ag >< '_key')));
    for s in il do
	unless isword(hd(s)) and isdeclared(hd(s)) and member(valof(hd(s)),csl) then
            mishap('Invalid slot: ' >< s >< ' for agent ' >< ag,[]);
        endunless;
    endfor;
enddefine;

define check_proc(pn,args,which);
	lvars p = valof(pn);
    unless isprocedure(p) then
		mishap('Need word identifier for ' >< which >< ': ',[^pn]);
    elseunless pdnargs(p) = args then
	    mishap(which >< ' procedure with ' >< args >< ' argument(s) needed: ',[^pn]);
	endunless;
enddefine;

;;; adds a contructor for a given type only if it is not present
define add_constructor_check(agtype);
    unless constructors(agtype) then
	;;; store the standard constructor
	valof(consword('new' >< agtype)) -> constructors(agtype);
    endunless;
enddefine;
	
/*
  The simulation now expects the following parameter format:

  [parameterlist entitylist resourcelist statslist]

  The "parameterlist" contains basic information about the simulation, such
  as run time, file names, output formats, number of decimals printed,
  etc.,

    [[output {SCREEN | FILE <filename> {ASCII | HTML}]
     [runtime {<num> | <termproc}]
     [decimals <num>]
     [setup <setupproc>]
    ]

events fromcycle tocycle <proc>... -- to allow particular changes to happen

  The <termproc> is a procedure that takes as its argument the number of
  simulation cycles to decide whether the simulation should end.  It needs
  to set true -> sim_stopping_scheduler; <setupproc> is a function that
  gets passed a list of all initially created agents and can be used to do
  more simulation specific setup

  The "entitylist" is a list of entity records for each entity to be
  created at the beginning, which specifies what to do with it

  Each entry has the following format

    [<entity>
       [<keyword> <item1> <item2> ...]
       [<kewword> <item1> <item2> ...]
       ...
    ]

  where <entity> is a defined (derived) sim_object class;

  The "startup" keyword defines the kinds of entities that should be
  created at the beginning of the simulation.  The following forms are
  available:

    startup <num> &<initlist>
    startup [<initlist1> &<initproc1>] [<initlist2> &<initproc2>] ...

  <num> is the number of agents

  <initlist> is a list [[<slot1> <val1>] [<slot2> <val2>] ...] where
  <slot1>,<slot2>, etc. are slots in the class <entity> and <val1>,<val2>,
  etc. the corresponding values

  <initproc> is a procedure that will be called with the agent passed as an
  argument after the initalizing the slots values with initlist (if
  present)

  The "initial" keyword defines what should happend upon creation of a new
  agent of type <entity>:

    initial &<initproc>

  <initproc> will be passed either two arguments, the newly created agent
  and its parent, if procreation is asexual, or a list of its parents, if
  procreation is sexual.  This function can be used to perform
  "inheritance" on certain slot values, for example, or "crossover".  If
  "initial" is omitted, the agent will be created with its default values.

  The "offspring" keyword defines whether agents of kind <entity> can turn
  into agents of kind <kind>.

    offspring [<transprob> <kind1> &<transproc1>] [<transprob> <kind2> &<transproc2>] ...

  <transprob> is the probability that an agent will transform into type
  <kind>, which can be any class derived from sim_object.  The optional
  transproc procedure takes as its two arguments the newly created agent of
  kind <kind> (which has only its default slots!!!), and the parent (of
  type <entity>).  When transformation takes place, no initialization
  occurs.

  The "mutate" keyword defines which slots should be mutated and how

    mutate [<mutprob1> [<slot1_1> <mutproc1_1>] [<slot1_2> <mutproc1_2>] ...]
           [<mutprob2> [<slot2_1> <mutproc2_1>] [<slot2_2> <mutproc2_2>] ...] ...

  where <mutprob> is the probability for mutation (upon creating of a new
  agent of type <entity>) of all the slots specified subsequenly by
  mutproc1, which is a procedure that takes two arguments, an agent and a
  slot.  Mutation always takes place after initialization and
  transformation.

  The "record" keyword is used to record statistical information about
  agents of type <entity>.  For every important event in an agent's life
  statistics can be recorded

    record [<freq1> [<slot1_1> &<recproc1_1>] [<slot2_1> &<recproc2_1>] ...]
           [<freq2> [<slot2_1> &<recproc2_1>] [<slot2_2> &<recproc2_2>] ...] ...

  The frequencies are for individual agents relative to their lifetime.
  <freqs> are either numbers, or alternatively "birth" or "death".  The
  optional <recpro> returns a value to be recorded, otherwise the slot
  value is taken instead.  All records will be saved at then end of an
  agent's lifetime and can be turned into statistics classes at the end of
  the simulation.

  The "resourcelist" defines regular activities that should take place at
  the beginning and the end of simulation cycles.  It is a list of any of
  the following entries:

    [always <proc1> <proc2> ...
    [at [<cycle1> [<kind1_1> &<initproc1_1>] [<kind1_2> &<initproc1_2>] ...]
        [<cycle2> [<kind2_1> &<initproc2_1>] [<kind2_2> &<initproc2_2>] ...] ...
    [every [<freq1> [<kind1_1> &<initproc1_1>] [<kind1_2> &<initproc1_2>] ...]
           [<freq2> [<kind2_1> &<initproc2_1>] [<kind2_2> &<initproc2_2>] ...] ...
    [random [<prob1> [<kind1_1> &<initproc1_1>] [<kind1_2> &<initproc1_2>] ...]
            [<prob2>  [<kind2_1> &<initproc2_1>] [<kind2_2> &<initproc2_2>] ...] ...

  <proc> takes a list of all sim_objects and a cycle number, <initproc>
  takes an object to initalize

  these procedures are called

    [record [<freq1> [<kind1_1> <op1_1> <slot1_1_1> <slot1_1_2> ...]
                     [<kind1_2> <op1_2> <slot1_2_1> <slot_1_2_2>] ...]
            [<freq2> [<kind2_1> <op2_1> <slot2_1_1> <slot2_1_2> ...]
                     [<kind2_2> <op2_1> <slot2_2_1> <slot_2_2_2>] ...] ...

  <frequency> is the rate at which stats are recorded for <kind>, <op>
  takes a list of agents and a list of slots as arguments and produces a
  result, which will be stored indexed by the cycle number and then by
  <kind>

  some basic numbers, e.g., the number of objects of each kind at any
  update cycle are automatically recorded... (alive_freq ???)

  SHOULD WE ALLOW FOR THE RECORDING OF DEAD AGENTS?

  The "statslist" identifies which statistics should be gathered:

    [statslist [<stat_list1>] [<stat_list2>] [<stat_list3>] ... ]

  For each statistic to be computed on the client, the following stat_list
  is needed:

    [<loc> [<stat1> <stat2> ... ] [<agent_type1> <agent_type2> ... ] <selector>
	[[<slot1> <conversion_func>] [<slot2> <conversion_func>] ... ]]

  where <loc> is the location of the computation, <statN> is the name of
  the statistic to be computed (e.g., avg, corr, val), agent_typeN is the
  simworld type of the agent to be examined, <selector> is one of (all |
  alive | dead), slotN is the name of an agent slot, and conversion_func is
  a function to convert categorical data to numerical.

  The server should be able to perform aggregate statistics on data
  returned from the clients (e.g., correlations between water efficiency
  and average survival).  Such a request is made with a <loc> of server.
  Client-side statistics use a <loc> of client.

  The client will send the results back in a similar list.  Note, however,
  that the above format allows several stats to be described for several
  agent types.  The results will need to be returned individually:

    [<loc> <stat> <agent_type> <selector> [<slot1> <slot2> ...] <value>]

  This identifies which statistic for which agent type for which selector
  group for which slots the value applies.  (An alternative that might be
  easier is for the server to provide each client stat with a unique ID
  number and send that along with the request, so that the client woudl
  only need to send an (ID, value) pair back to the server.  This would
  eliminate a little bit of parsing.)

*/

define parseentitylist(entitylist,restypes) -> (alive_agents,agenttypes);
    ;;; create the respective objects, and initalize statistics
    ;;; NOTE: this assumes that every agent type is only mentioned ONCE!!!
    lvars agenttypes = restypes, alive_agents = [], kw, ag;

    [] -> dead_agents;
    ;;; first make sure that all the constructors for all entity types
    ;;; mentioned in the resource list are there
    for ag in restypes do add_constructor_check(ag); endfor;

    returnif(null(entitylist));
    ;;; create alive_agents list
    [% for ag in entitylist do
	lvars agtype= if isword(hd(ag)) then hd(ag) else consword(hd(ag)) endif;
    	
	;;; check if constructor exists, which it should if the GUI class
	;;; has been defined
	;;; MS: DO WE NEED TO DO THIS AGAIN?
	add_constructor_check(agtype);
	
	;;; check if the agenttype has been used before and indicate error
	if lmember(agtype,agenttypes) then
	    if verbose then
		npr('Duplicate agent kind entry in entity list: ' >< agtype);
	    endif;
	else
		conspair(agtype,agenttypes) -> agenttypes;
	endif;

	;;; now examine keywords
	for kw in tl(ag) do
	    lvars k = hd(kw);
	    if k == "startup" then
		;;; startup <num> &<initlist>
    		;;; startup [<initlist1> &<initproc1>] [<initlist2> &<initproc2>] ...
		if isnumber(kw(2)) then
		    lvars initlist = tl(tl(kw));
		    if initlist /= [] then
			hd(initlist) -> initlist;
		    endif;
		    ;;; check if slots are defined for agent kind
		    check_initlist(initlist,agtype);
		    repeat kw(2) times
			lvars obj = newobject(constructors,[],agtype,initlist);
			obj
		    endrepeat
 		else
		    lvars ent;
		    for ent in tl(kw) do
			lvars obj = newobject(constructors,[],agtype,hd(ent));
			;;; check if there is an initproc and if so, run it
			unless null(tl(ent)) then
			    check_proc(ent(2),1,'Startup');
			    valof(ent(2))(obj);
			endunless;
			obj
		    endfor;
		endif;
	    elseif k == "initial" then
		;;; initial <initproc>, check if second entry is a
		;;; procedure word identifier
		check_proc(kw(2),2,'Initialization');
		;;; override method definition for agtype
		compile(stringin('define :method initialize(obj:' >< agtype
		    >< ',parents);\n'><kw(2)><'(obj,parents);\nenddefine;'));
	    elseif k == "offspring" then
    		;;; offspring [<transprob> <kind1> &<transproc1>] [<transprob> <kind2> &<transproc2>] ...
		;;; this orders the transformations implicitly by giving
		;;; higher priority to the initials ones hence, low
		;;; probabilities should probably come first
		lvars trans, str = 'define :method offspring(obj:'
		    >< agtype><') -> newthing;\n';
		for trans in tl(kw) do
		    ;;; PWS call to newobject needs to be fixed for passing
		    ;;; offspring
		    str >< 'if random(1.0)<=' >< trans(1)
			><' then\nnewobject(constructors,"' >< trans(2)
			>< '")->newthing;\n' -> str;
		    ;;; make sure the constructor is in the list...
		    add_constructor_check(trans(2));
		    ;;; check if in GUI mode, and if so add windowlist to
		    ;;; agent and show agent in windows check if there is a
		    ;;; transformation procedure that should be applied
		    if length(trans)>2 then
			check_proc(trans(3),2,'Transformation');
 			str >< trans(3) ><'(newthing,obj);\n'->str;
		    endif;
                    str >< 'else' -> str;
		endfor;
		;;; if no transformation takes place then produce an
		;;; offspring of the parent kind and initialize it
		;;; PWS call to newobject needs to be fixed for passing
		;;; offspring
		str >< '\nnewobject(constructors,"'><agtype
		    ><'") -> newthing;\nendif;\nenddefine;' -> str;
		;;;str >< '\nnewobject(constructors,"'><agtype><'") -> newthing;initialize(newthing,obj);\nendif;\nenddefine;' -> str;
		;;; now compile it
		compile(stringin(str));
	    elseif k == "mutate" then
		;;; mutate [<mutprob1> [<slot1_1> <mutproc1_1>] [<slot1_2> <mutproc1_2>] ...]
		;;;        [<mutprob2> [<slot2_1> <mutproc2_1>] [<slot2_2> <mutproc2_2>] ...] ...
		;;; instead of <mutproc> an interval from which a
		;;; random number will be generated is also possible
		;;; the interval is given as a list [lower upper]
		lvars mut, ent, str = 'define :method mutate(obj:'
		    ><agtype><');\n';
		for mut in tl(kw) do
		    str >< 'if random(1.0)<=' >< mut(1) ><' then\n' -> str;
		    for ent in tl(mut) do
			if islist(ent(2)) then
			    lvars inter = ent(2),stretch =
				  abs(inter(2)-inter(1));
			    str >< inter(1) >< ' + random(' >< stretch
				>< ') -> ' >< ent(1) >< '(obj);\n' -> str;
			else
			    check_proc(ent(2),2,'Mutation');
			    str >< ent(2) >< '(obj,'>< ent(1)>< ' -> '
				>< ent(1) >< '(obj);\n' -> str;
			endif;
		    endfor;
		    str >< 'endif;\n' -> str;
		endfor;
		str >< 'enddefine;' -> str;
		;;; now compile it
		compile(stringin(str));		
	    elseif k == "record" then
		;;; record [every <freq1> [<slot1_1> &<recproc1_1>] [<slot2_1> &<recproc2_1>] ...]
		;;;    [at <freq2> [<slot2_1> &<recproc2_1>] [<slot2_2> &<recproc2_2>] ...] ...
		;;; stats will be saved in properties of properties, one for each agent which is produced automatically
		lvars rec, ent, str = 'define :method record(obj:'
		    ><agtype><',cycle);\n';
		str >< 'unless RECINDSLOTS(sim_name(obj)) then '
		    >< 'consvector(repeat ' >< length(tl(kw))
		    >< ' times [] endrepeat,' >< length(tl(kw))
		    >< ') -> RECINDSLOTS(sim_name(obj));\nendunless;\n'
		    -> str;
		str >< 'lvars vec = RECINDSLOTS(sim_name(obj));' -> str;
		lvars which = 1;
		for rec in tl(kw) do
		    if rec(1) == "every" then
			str >< 'if cycle rem ' >< rec(2) ><' = 0 then\n'
			    -> str;
		    elseif rec(1) == "at" then
			if rec(2) == "birth" then
			    str >< 'if age(obj) = 0 then\n' -> str;
			elseif rec(2) == "death" then
			    str >< 'if sim_status(obj) = "dead" then \n'
				- str;
			else
			    str >< 'if age(obj) = ' >< rec(2) ><' then\n'
				-> str;
			endif;
		    else
			mishap('Unkown recording option: ',[^(rec(1))]);
		    endif;
		    str >< 'conspair([% cycle, ' -> str;
		    for ent in tl(tl(rec)) do
			unless null(tl(ent)) then
			    check_proc(ent(2),2,'Record');
			    str >< ent(2) >< '(obj,'>< ent(1) ><'),' -> str;
			;;; no procedure specified, take slot values instead
			else
			    str >< ent(1) >< '(obj),' -> str;
			endunless;
		    endfor;
		    str >< '%],vec(' >< which >< ')) -> vec(' >< which
			>< ');\nendif;\n' -> str;
		    which + 1 -> which;
		endfor;
		str >< 'enddefine;' -> str;
		;;; now compile it
		compile(stringin(str));		   	
	    else
		unless ignoreUNKNOWNKEYWORD then
		    mishap('Unkown keyword in parameter list: ',[^k]);
		endunless;
	    endif;
	 endfor;
	endfor %] -> alive_agents;
enddefine;

/*
    [always <proc1> <proc2> ...
    [every [<freq1> [<kind1_1> &<initproc1_1>] [<kind1_2> &<initproc1_2>] ...]
           [<freq2> [<kind2_1> &<initproc2_1>] [<kind2_2> &<initproc2_2>] ...] ...
    [random [<prob1> [<kind1_1> &<initproc1_1>] [<kind1_2> &<initproc1_2>] ...]
            [<prob2>  [<kind2_1> &<initproc2_1>] [<kind2_2> &<initproc2_2>] ...] ...

  <proc> takes a list of all sim_objects and a cycle number, <initproc>
  takes an object to initalize, <termproc> takes an object to check for
  termination of the object
*/

;;; parses the resource list and returns the new types of agents encountered
define parseresourcelist(resourcelist)->agenttypes;
    lvars agenttypes = [], kw, ag;

    compile(stringin('define update_resources(sim_objects,sim_cycle_number);enddefine;'));
    returnif(null(resourcelist));
    ;;; assemble definitionstring for update_resource function
    ;;; this produces a list of strings, which are all concatenated and compiled
    compile(stringin(applist('define update_resources(sim_objects,'
	>< 'sim_cycle_number);\nlvars obj = [];\n',
	[% ;;; examine keywords
	    for kw in resourcelist do
		lvars k = hd(kw);
		if k == "always" then
		    lvars proc;
		    for proc in kw(2) do
		    	proc >< '(sim_objects,sim_cycle_number);\n'
		    endfor;
		elseif k == "every" then
		    lvars proc, ent;
		    for proc in tl(kw) do
			'if sim_cycle_number rem ' >< proc(1) ><' = 0 then\n';
			    if islist(hd(tl(proc))) then
				npr('Found pair');
				for ent in tl(proc) do
				    'conspair(newobject(constructors,[],"'
					>< ent(1) >< '"),obj)->obj;\n';
				    ;;; check if ent(1) has a constructor
				    add_constructor_check(ent(1));
				    unless member(ent(1),agenttypes) then
					conspair(ent(1),agenttypes) -> agenttypes;
				    endunless;
				    if null(tl(ent)) then
					'initialize'
				    else
					ent(2)
				    endif
				    >< '(hd(obj),[]);\n';
				endfor;
			    else
				lvars proc;
				for ent in tl(proc) do
				    ent >< '(sim_objects,sim_cycle_number);\n'
				endfor;
			    endif;
			'endif;\n';
		    endfor;
		elseif k == "at" then
		    lvars proc, ent;
		    for proc in tl(kw) do
			'if sim_cycle_number = ' >< proc(1) ><' then\n';
			    for ent in tl(proc) do
				'conspair(newobject(constructors,[],"'
				    >< ent(1) >< '"),obj)->obj;\n';
				;;; check if ent(1) has a constructor
				add_constructor_check(ent(1));
				unless member(ent(1),agenttypes) then
				    conspair(ent(1),agenttypes) -> agenttypes;
				endunless;
				if null(tl(ent)) then
			    	    'initialize'
				else
			    	    ent(2)
				endif
				>< '(hd(obj),[]);\n';
			    endfor;
			'endif;\n';
		    endfor;
		elseif k == "random" then
		    lvars proc, ent;
		    for proc in tl(kw) do
			'if random(1.0)<= ' >< proc(1) ><' then\n';
			    for ent in tl(proc) do
				'conspair(newobject(constructors,[],"'
				    >< ent(1) >< '"),obj)->obj;\n';
				;;; check if ent(1) has a constructor
				add_constructor_check(ent(1));
				unless member(ent(1),agenttypes) then
				    conspair(ent(1),agenttypes) -> agenttypes;
				endunless;
				if null(tl(ent)) then
				    'initialize'
				else
			    	    ent(2)
				endif
				>< '(hd(obj),[]);\n';
			    endfor;
			'endif;\n';
		    endfor;
	    ;;;[record [<freq1> [<kind1_1> <op1_1> <slot1_1_1> <slot1_1_2> ...] [<kind1_2> <op1_2> <slot1_2_1> <slot_1_2_2>] ...]
        ;;;        [<freq2> [<kind2_1> <op2_1> <slot2_1_1> <slot2_1_2> ...] [<kind2_2> <op2_1> <slot2_2_1> <slot_2_2_2>] ...] ..
		;;; PWS: does this belong here?
		elseif k == "record" then
		    ;;; sort out how many agent kinds there are, then
		    ;;; create lists of agents for each agent kind and pass
		    ;;; these lists to the functions...

		else
 		    unless ignoreUNKNOWNKEYWORD then
	    		mishap('Unkown keyword in parameter list: ',[^k]);
	    	    endunless;
		endif;
	    endfor;
	    'sim_object_add_list <> obj -> sim_object_add_list;\nenddefine;' %],
	    procedure(x,y);x >< y;endprocedure)));
enddefine;

/* for now this is the format:

[[output {SCREEN | NONE| FILE <filename>]
 [format {ASCII | HTML}]
 [quitif {<num> | <termproc}]
 [decimals <num>]
 [setup <setupproc>]
 [seed <num>]
]
*/
define parseparams(parameterlist)->(file,printmode,runtime);
    ;;; default values
    lvars kw, file = [], printmode = "ASCII", runtime = 1000;

    for kw in parameterlist do
	lvars k = hd(kw);
	if k == "output" then
	    if kw(2) == "SCREEN" then
		[] -> file;
	    elseif kw(2) == "FILE" then
		kw(3) -> file;
	    elseif kw(2) == "NONE" then
		[] -> file;
		"NONE" -> printmode;
	    else
		mishap('Unknown output device: ',[^(kw(2))]);
	    endif;
	elseif k == "format" then
	    if kw(2) /== "ASCII" and kw(2) /== "HTML" then
		mishap('Unkown output format: ',[^kw(2)]);
	    endif;
	    kw(2) -> printmode;
	elseif k == "quitif" then
	    if not(isnumber(kw(2))) and not(isprocedure(kw(2))) then
		mishap('Unsupported termination conditiion: ',[^(kw(2))]);
	    endif;
	    kw(2) -> runtime;
	elseif k == "seed" then
	    if not(isnumber(kw(2))) then
		mishap('Number required for seed: ',[^(kw(2))]);
	    endif;
            kw(2) -> ranseed;
	elseif k == "decimals" then
	    if not(isnumber(kw(2))) then
		mishap('Number required for decimals: ',[^(kw(2))]);
	    endif;
            kw(2) -> sim_pr_places;
	elseif k == "setup" then
	    unless isprocedure(kw(2)) then
		mishap('Procedure required for setup: ',[^(kw(2))]);
            endunless;
	    ;;; assemble and compile it
	    compile(stringin('define finish_setup(alive_agents);\n' >< kw(2) ><
                '(alive_agents);\nenddefine;'));
	else
	    unless ignoreUNKNOWNKEYWORD then
	    	mishap('Unkown keyword in parameter list: ',[^k]);
	    endunless;
	endif;
    endfor;
enddefine;

;;; the main simulator loop
define simulate(params);
    lvars (parameterlist,entitylist,resourcelist) = explode(params),
    	 ;;; parse the simulation parameters
    	 (file, printmode, runtime) = parseparams(parameterlist);

    ;;; parse the resources and set them up, and the entities and create them
    parseentitylist(entitylist,parseresourcelist(resourcelist))
		-> (alive_agents,agenttypes);

    ;;; save the random seed
    ranseed -> originalseed;

    ;;; clear the gensym property for agent names...
    clearproperty(gensym_property);

    ;;; print results, to file if it exists...
    dlocal cucharout;
    global vars oldcucharout = cucharout;

    ;;; check if it needs to be redirected to a file
    ;;; output format is either HTML or ASCII
    if (file /= []) then
    	lvars htmlfile = file >< if printmode == "HTML" then '.html'
	     elseif printmode == "ASCII" then '.txt' else '' endif;
        ;;; re-direct print output to file
        if sys_file_exists(htmlfile) then
            ;;; delete it if it exists
            sysdelete(htmlfile);
        endif;
        discout(htmlfile) -> cucharout;
    endif;

    ;;; print out simulation setup
    if printmode == "HTML" then
        ;;; print the start up data
        npr('<html><head><meta name="Author" content="Matthias Scheutz">' ><
            '<title>Results</title>' >< '</head><body>');
        printf(originalseed,line_repeater(pipein('/bin/date',[],false),255)(),
            '<center>Simulation started: %p with seed %p</center><br>\n');
        printf('<h2>Simulation Parameters</h2>\n');
        printlist(params,printmode);
    elseif printmode = "ASCII" then
        printf(originalseed,line_repeater(pipein('/bin/date',[],false),255)(),
            'Simulation started: %p with seed %p\n');
        printf('Simulation Parameters\n');
        printlist(params,printmode);
	npr('');
    endif;

    ;;; if output went to a file, close it
    if (file /= []) then
        cucharout(termin);
	;;; restore charrepeater
 	oldcucharout -> cucharout;
    endif;

    ;;; clear all single run statistics (actually only for GUI mode)

    clearproperty(alive_freq);
    clearproperty(agent_alive);
    clearproperty(agent_age);
    clearproperty(agent_hunger);
    clearproperty(agent_eaten);
    clearproperty(agent_crash);
    clearproperty(agent_thirst);
    clearproperty(agent_energy);
    clearproperty(agent_water);
    clearproperty(agent_distance);
    clearproperty(agent_offspring);
    clearproperty(agent_eef);
    clearproperty(agent_wef);
    clearproperty(RECINDSLOTS);

    ;;; set the startup numbers for each agent type
    lvars typ,a;
    for typ in agenttypes do
	lvars quant = 0;
	for a in alive_agents do
	    if class_name(datakey(a)) = typ then
		quant + 1 -> quant;
	    endif;
	endfor;
    	conslist(quant,1) -> alive_freq(typ);
    endfor;

    ;;; Specify decimal places when printing numbers
    dlocal pop_pr_places = sim_pr_places;

    ;;; set simworld tracing to false
    lvars word;
    for word in sim_alltracevars do false -> valof(word) endfor;

    ;;; call the user-definable setup function passing all agents as arguments
    finish_setup(alive_agents);
	
    ;;; start the sim_agent_scheduler and run the simulation
    sim_scheduler_objects(alive_agents,runtime) -> alive_agents;

    ;;; if we are writing to a file, open it for appending
    if (file /= []) then
	lvars htmlfile = file >< if printmode == "HTML" then '.html'
	     elseif printmode == "ASCII" then '.txt' else '' endif;
        discappend(htmlfile) -> cucharout;
    endif;

    if printmode == "HTML" then
       	;;; print out some statistics in HTML
       	printagentstats(alive_agents,agenttypes,"HTML");
       	printf(line_repeater(pipein('/bin/date',[],false),255)(),
               '<hr><center>Created: %p<br>(c) Matthias Scheutz</center>');
        npr('</body></html>');
    elseif printmode == "ASCII" then
	printagentstats(alive_agents,agenttypes,"ASCII");
        printf(line_repeater(pipein('/bin/date',[],false),255)(),
            'Simulation completed! %p\n');
    else
	;;; just compute them (used in client mode)
	printagentstats(alive_agents,agenttypes,false);
    endif;

    ;;; close the file if it was open
    if (file /= []) then
        cucharout(termin);
    	;;; and create stats file if needed
    	;;; change file mode to -rw-r--r--
    	;;;33188 -> sysfilemode(htmlfile);
    	;;; no print extended stats in separate file
    	file >< 'stat.txt' -> file;
    	;;; re-direct print output to file
        if sys_file_exists(file) then
            ;;; delete it
            sysdelete(file);
        endif;
        discout(file) -> cucharout;
    endif;

    ;;; add the agents that are alive to the stats and
    ;;; mark them as alive
    lvars item;
    for item in alive_agents do
	if isbasic_agent(item) then
	    agent_age(kind(item)) + age(item) -> agent_age(kind(item));
	endif;
    endfor;

    ;;; reverse the list (was built in reverse order for speed reasons)
    for item in agenttypes do
        rev(alive_freq(item))->alive_freq(item);
    endfor;

    ;;; print the alive frequencies if needed
    lvars i,j;
    if printmode == "HTML" or printmode == "ASCII" then
        nprintf('\nPrinting alive frequencies for ' >< agenttypes >< ' total');
        for i from 1 to length(alive_freq(agenttypes(1))) do
            lvars tot = 0,j;
            for j from 1 to length(agenttypes) do
                printf('%p, ',[%alive_freq(agenttypes(j))(i)%]);
            endfor;
            npr('');
        endfor;
        ;;; change file mode to -rw-r--r--
        ;;;33188 -> sysfilemode(file);
    endif;
    if (file /= []) then
        cucharout(termin);
    endif;
enddefine;

/*
;;; test-call
simulate([
	  [[quitif 1000]]
	
	  [[basic_agent
	    [startup 20]
	    [mutate [0.15 [food_factor [-100 100]]]
	     [0.15 [water_factor [-100 100]]]
	     [0.15 [obst_factor [-100 100]]]
	     [0.15 [agent_factor [-100 100]]]]
	    ]
	  [obstacle
	   [startup 1]
	   ]]
	
 	  [[random
	    [0.25 [plant]]
	    [0.15 [water]]
	    ]]
	  ]);
*/
